
from xml.dom.minidom import Element

from pyutmodel.PyutClassCommon import PyutClassCommon

from oglio.toXmlV10.BaseToDom import BaseToDom
from oglio.toXmlV10.XmlConstants import XmlConstants


class BasePyutToDom(BaseToDom):

    def __init__(self):
        super().__init__()

    def _pyutClassCommonToXml(self, classCommon: PyutClassCommon, root: Element) -> Element:

        root.setAttribute(XmlConstants.ATTR_DESCRIPTION, classCommon.description)
        # root.setAttribute(PyutXmlConstants.ATTR_FILENAME,    pyutInterface.getFilename())

        return root



from typing import Dict
from typing import List
from typing import NewType
from typing import Union
from typing import cast

from dataclasses import dataclass
from dataclasses import field

from ogl.OglActor import OglActor
from ogl.OglNote import OglNote
from ogl.OglClass import OglClass
from ogl.OglLink import OglLink
from ogl.OglText import OglText
from ogl.OglUseCase import OglUseCase
from ogl.sd.OglSDInstance import OglSDInstance
from ogl.sd.OglSDMessage import OglSDMessage

from untanglepyut.UnTangler import Document
from untanglepyut.UnTangler import ProjectInformation

OglClasses  = NewType('OglClasses',  List[OglClass])
OglLinks    = NewType('OglLinks',    List[OglLink])
OglNotes    = NewType('OglNotes',    List[OglNote])
OglTexts    = NewType('OglTexts',    List[OglText])
OglActors   = NewType('OglActors',   List[OglActor])
OglUseCases = NewType('OglUseCases', List[OglUseCase])
OglSDInstances = NewType('OglSDInstances', Dict[int, OglSDInstance])
OglSDMessages  = NewType('OglSDMessages',  Dict[int, OglSDMessage])

OglObjects = Union[OglClasses, OglLinks, OglNotes, OglTexts, OglActors, OglUseCases]


def createOglClassesFactory() -> OglClasses:
    """
    Factory method to create  the OglClasses data structure;

    Returns:  A new data structure
    """
    return OglClasses([])


def createOglLinksFactory() -> OglLinks:
    """
    Factory method to create  the OglLinks data structure;

    Returns:  A new data structure
    """
    return OglLinks([])


def createOglNotesFactory() -> OglNotes:
    return OglNotes([])


def createOglTextsFactory() -> OglTexts:
    return OglTexts([])


def createOglActorsFactory() -> OglActors:
    return OglActors([])


def createOglUseCasesFactory() -> OglUseCases:
    return OglUseCases([])


def createOglSDInstances() -> OglSDInstances:
    return OglSDInstances({})


def createOglSDMessages() -> OglSDMessages:
    return OglSDMessages({})


OglDocumentTitle = NewType('OglDocumentTitle', str)


@dataclass
class OglDocument:
    documentType:    str = ''
    documentTitle:   OglDocumentTitle = OglDocumentTitle('')
    scrollPositionX: int = -1
    scrollPositionY: int = -1
    pixelsPerUnitX:  int = -1
    pixelsPerUnitY:  int = -1
    oglClasses:      OglClasses  = field(default_factory=createOglClassesFactory)
    oglLinks:        OglLinks    = field(default_factory=createOglLinksFactory)
    oglNotes:        OglNotes    = field(default_factory=createOglNotesFactory)
    oglTexts:        OglTexts    = field(default_factory=createOglTextsFactory)
    oglActors:       OglActors   = field(default_factory=createOglActorsFactory)
    oglUseCases:     OglUseCases = field(default_factory=createOglUseCasesFactory)
    oglSDInstances:  OglSDInstances = field(default_factory=createOglSDInstances)
    oglSDMessages:   OglSDMessages  = field(default_factory=createOglSDMessages)

    def toOglDocument(self, document: Document):
        self.documentType    = document.documentType
        self.documentTitle   = OglDocumentTitle(document.documentTitle)
        self.scrollPositionX = document.scrollPositionX
        self.scrollPositionY = document.scrollPositionY
        self.pixelsPerUnitX  = document.pixelsPerUnitX
        self.pixelsPerUnitY  = document.pixelsPerUnitY


OglDocuments     = NewType('OglDocuments', dict[OglDocumentTitle, OglDocument])


def createOglDocumentsFactory() -> OglDocuments:
    return OglDocuments({})


@dataclass
class OglProject:
    fileName: str = cast(str, None)
    version:  str = cast(str, None)
    codePath: str = cast(str, None)
    oglDocuments: OglDocuments = field(default_factory=createOglDocumentsFactory)

    def toOglProject(self, project: ProjectInformation):
        self.fileName = project.fileName
        self.version  = project.version
        self.codePath = project.codePath




from logging import Logger
from logging import getLogger

from xml.dom.minidom import Document
from xml.dom.minidom import Element

from miniogl.AttachmentSide import AttachmentSide
from miniogl.SelectAnchorPoint import SelectAnchorPoint

from ogl.OglAssociation import OglAssociation
from ogl.OglAssociationLabel import OglAssociationLabel
from ogl.OglInterface2 import OglInterface2
from ogl.OglLink import OglLink

from oglio.Types import OglLinks
from oglio.toXmlV10.BaseOglToDom import BaseOglToDom
from oglio.toXmlV10.PyutToDom import PyutToDom
from oglio.toXmlV10.XmlConstants import XmlConstants


class OglLinksToDom(BaseOglToDom):

    def __init__(self, xmlDocument: Document):

        super().__init__(xmlDocument=xmlDocument)
        self.linksLogger: Logger = getLogger(__name__)

        self._pyutToMiniDom: PyutToDom = PyutToDom()

    def serialize(self, documentNode: Element, oglLinks: OglLinks) -> Element:

        for oglLink in oglLinks:
            if isinstance(oglLink, OglInterface2):
                lollipopElement: Element = self._oglInterface2ToDom(oglLink, self._xmlDocument)
                documentNode.appendChild(lollipopElement)
            else:
                linkElement: Element = self._oglLinkToDom(oglLink=oglLink, xmlDoc=self._xmlDocument)
                documentNode.appendChild(linkElement)

        return documentNode

    def _oglInterface2ToDom(self, oglInterface: OglInterface2, xmlDoc: Document) -> Element:
        """

        Args:
            oglInterface:   Lollipop to convert
            xmlDoc:         xml document

        Returns:
            New minidom element
        """
        root: Element = xmlDoc.createElement(XmlConstants.ELEMENT_GRAPHIC_LOLLIPOP)

        destAnchor:      SelectAnchorPoint = oglInterface.destinationAnchor
        attachmentPoint: AttachmentSide   = destAnchor.attachmentPoint
        x, y = destAnchor.GetPosition()

        root.setAttribute(XmlConstants.ATTR_LOLLIPOP_ATTACHMENT_POINT, attachmentPoint.__str__())
        root.setAttribute(XmlConstants.ATTR_X, str(x))
        root.setAttribute(XmlConstants.ATTR_Y, str(y))

        # parentUmlClass: OglClass = destAnchor.GetParent()
        # parentId:       int      = self._idFactory.getID(parentUmlClass.getPyutObject())
        # self.logger.info(f'Interface implemented by class id: {parentId}')

        # root.setAttribute(PyutXmlConstants.ATTR_IMPLEMENTED_BY_CLASS_ID, str(parentId))
        root.appendChild(self._pyutToMiniDom.pyutInterfaceToDom(oglInterface.pyutInterface, xmlDoc))

        return root

    def _oglLinkToDom(self, oglLink: OglLink, xmlDoc: Document):
        """
        Export an OgLink to a minidom element
        Args:
            oglLink:    OglLink to convert
            xmlDoc:     xml document

        Returns:
            A new minidom element
        """
        root = xmlDoc.createElement(XmlConstants.ELEMENT_GRAPHIC_LINK)

        # save source and destination anchor points
        x, y = oglLink.sourceAnchor.GetModel().GetPosition()
        simpleX, simpleY = self._getSimpleCoordinates(x, y)
        root.setAttribute(XmlConstants.ATTR_LINK_SOURCE_ANCHOR_X, simpleX)
        root.setAttribute(XmlConstants.ATTR_LINK_SOURCE_ANCHOR_Y, simpleY)

        x, y = oglLink.destinationAnchor.GetModel().GetPosition()
        simpleX, simpleY = self._getSimpleCoordinates(x, y)

        root.setAttribute(XmlConstants.ATTR_LINK_DESTINATION_ANCHOR_X, simpleX)
        root.setAttribute(XmlConstants.ATTR_LINK_DESTINATION_ANCHOR_Y, simpleY)

        root.setAttribute(XmlConstants.ATTR_SPLINE, str(oglLink.GetSpline()))

        if isinstance(oglLink, OglAssociation):

            center: OglAssociationLabel = oglLink.centerLabel
            src:    OglAssociationLabel = oglLink.sourceCardinality
            dst:    OglAssociationLabel = oglLink.destinationCardinality

            assocLabels = {
                XmlConstants.ELEMENT_ASSOC_CENTER_LABEL:      center,
                XmlConstants.ELEMENT_ASSOC_SOURCE_LABEL:      src,
                XmlConstants.ELEMENT_ASSOC_DESTINATION_LABEL: dst
            }
            for eltName in assocLabels:
                elt: Element = self._createAssocLabelElement(eltName, xmlDoc, assocLabels[eltName])
                root.appendChild(elt)

        # save control points (not anchors!)
        for x, y in oglLink.segments[1:-1]:
            item = xmlDoc.createElement(XmlConstants.ELEMENT_MODEL_CONTROL_POINT)
            item.setAttribute(XmlConstants.ATTR_X, str(x))
            item.setAttribute(XmlConstants.ATTR_Y, str(y))
            root.appendChild(item)

        # adding the data layer object

        root.appendChild(self._pyutToMiniDom.pyutLinkToDom(oglLink.pyutObject, xmlDoc))

        return root

    # noinspection SpellCheckingInspection
    def _createAssocLabelElement(self, eltText: str, xmlDoc: Document, oglLabel: OglAssociationLabel) -> Element:
        """
        Creates an element of the form:

        ```html
        `<eltText x="nnnn" y="nnnn"/>`
        ```

        e.g.

        ```html
            `<LabelCenter x="1811" y="1137"/>`
        ```

        Args:
            eltText:    The element name
            xmlDoc:     The minidom document
            oglLabel:   A description of a label includes text and position

        Returns:
            A new minidom element
        """
        label: Element = xmlDoc.createElement(eltText)

        x, y = oglLabel.GetPosition()

        relativeX, relativeY = oglLabel.ConvertCoordToRelative(x=x, y=y)
        self.linksLogger.warning(f'x,y = ({x},{y})   relativeX,relativeY = ({relativeX},{relativeY})')
        label.setAttribute(XmlConstants.ATTR_X, str(relativeX))
        label.setAttribute(XmlConstants.ATTR_Y, str(relativeY))

        return label




class XmlConstants:
    """
    A `no method` class that just hosts the strings that represent the Pyut XML strings
    """

    TOP_LEVEL_ELEMENT:           str = 'PyutProject'
    ELEMENT_GRAPHIC_CLASS:       str = 'GraphicClass'
    ELEMENT_GRAPHIC_LOLLIPOP:    str = 'GraphicLollipop'
    ELEMENT_GRAPHIC_NOTE:        str = 'GraphicNote'
    ELEMENT_GRAPHIC_TEXT:        str = 'GraphicText'
    ELEMENT_GRAPHIC_ACTOR:       str = 'GraphicActor'
    ELEMENT_GRAPHIC_USE_CASE:    str = 'GraphicUseCase'
    ELEMENT_GRAPHIC_LINK:        str = 'GraphicLink'
    ELEMENT_GRAPHIC_SD_INSTANCE: str = 'GraphicSDInstance'
    ELEMENT_GRAPHIC_SD_MESSAGE:  str = 'GraphicSDMessage'

    ELEMENT_DOCUMENT:          str = 'PyutDocument'
    ELEMENT_MODEL_CLASS:       str = 'Class'
    ELEMENT_MODEL_INTERFACE:   str = 'Interface'
    ELEMENT_IMPLEMENTOR:       str = 'Implementor'
    ELEMENT_MODEL_METHOD:      str = 'Method'
    ELEMENT_MODEL_NOTE:        str = 'Note'
    ELEMENT_MODEL_TEXT:        str = 'Text'
    ELEMENT_MODEL_ACTOR:       str = 'Actor'
    ELEMENT_MODEL_USE_CASE:    str = 'UseCase'
    ELEMENT_MODEL_LINK:        str = 'Link'
    ELEMENT_MODEL_FIELD:       str = 'Field'
    ELEMENT_MODEL_PARAM:       str = 'Param'
    ELEMENT_MODEL_RETURN:      str = 'Return'
    ELEMENT_MODEL_MODIFIER:    str = 'Modifier'
    ELEMENT_MODEL_SOURCE_CODE: str = 'SourceCode'
    ELEMENT_MODEL_CODE:        str = 'Code'

    ELEMENT_MODEL_SD_INSTANCE:   str = 'SDInstance'
    ELEMENT_MODEL_SD_MESSAGE:    str = 'SDMessage'
    ELEMENT_MODEL_CONTROL_POINT: str = 'ControlPoint'

    ELEMENT_ASSOC_CENTER_LABEL:      str = 'LabelCenter'
    ELEMENT_ASSOC_SOURCE_LABEL:      str = 'LabelSrc'
    ELEMENT_ASSOC_DESTINATION_LABEL: str = 'LabelDst'

    ATTR_VERSION: str = 'version'

    ATTR_ID: str = 'id'

    ATTR_WIDTH:  str = 'width'
    ATTR_HEIGHT: str = 'height'

    ATTR_X: str = 'x'
    ATTR_Y: str = 'y'

    ATTR_DISPLAY_PARAMETERS: str = 'displayParameters'

    ATTR_STEREOTYPE:  str = 'stereotype'
    ATTR_DESCRIPTION: str = 'description'
    ATTR_VISIBILITY:  str = 'visibility'

    ATTR_FILENAME: str = 'filename'
    ATTR_NAME:     str = 'name'
    ATTR_CONTENT:  str = 'content'
    ATTR_TYPE:     str = 'type'
    ATTR_TITLE:    str = 'title'

    ATTR_DEFAULT_VALUE:   str = 'defaultValue'
    ATTR_SHOW_STEREOTYPE: str = 'showStereotype'
    ATTR_SHOW_METHODS:    str = 'showMethods'
    ATTR_SHOW_FIELDS:     str = 'showFields'

    ATTR_LINK_SOURCE_ANCHOR_X: str = 'srcX'
    ATTR_LINK_SOURCE_ANCHOR_Y: str = 'srcY'

    ATTR_LINK_DESTINATION_ANCHOR_X: str = 'dstX'
    ATTR_LINK_DESTINATION_ANCHOR_Y: str = 'dstY'

    ATTR_SPLINE:        str = 'spline'
    ATTR_BIDIRECTIONAL: str = 'bidir'

    ATTR_SOURCE_ID:      str = 'sourceId'
    ATTR_DESTINATION_ID: str = 'destId'

    ATTR_CARDINALITY_SOURCE:      str = 'cardSrc'
    ATTR_CARDINALITY_DESTINATION: str = 'cardDestination'

    ATTR_INSTANCE_NAME:    str = 'instanceName'
    ATTR_LIFE_LINE_LENGTH: str = 'lifeLineLength'

    ATTR_MESSAGE:               str = 'message'
    ATTR_SOURCE_TIME_LINE:      str = 'srcTime'
    ATTR_DESTINATION_TIME_LINE: str = 'dstTime'

    ATTR_SD_MESSAGE_SOURCE_ID:      str = 'srcID'
    ATTR_SD_MESSAGE_DESTINATION_ID: str = 'dstID'

    V9_LINK_PREFIX: str = 'OGL_'

    ATTR_CODE_PATH: str = 'CodePath'

    ATTR_SCROLL_POSITION_X: str = 'scrollPositionX'
    ATTR_SCROLL_POSITION_Y: str = 'scrollPositionY'
    ATTR_PIXELS_PER_UNIT_X: str = 'pixelsPerUnitX'
    ATTR_PIXELS_PER_UNIT_Y: str = 'pixelsPerUnitY'

    ATTR_LOLLIPOP_ATTACHMENT_POINT: str = 'attachmentPoint'
    ATTR_IMPLEMENTING_CLASS_NAME:   str = 'implementingClassName'

    ATTR_TEXT_SIZE:     str = 'textSize'
    ATTR_IS_BOLD:       str = 'isBold'
    ATTR_IS_ITALICIZED: str = 'isItalicized'
    ATTR_FONT_FAMILY:   str = 'fontFamily'




from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from xml.dom.minidom import Document
from xml.dom.minidom import Element

from oglio.Types import OglActors
from oglio.Types import OglClasses
from oglio.Types import OglDocument
from oglio.Types import OglLinks
from oglio.Types import OglNotes
from oglio.Types import OglSDInstances
from oglio.Types import OglSDMessages
from oglio.Types import OglTexts
from oglio.Types import OglUseCases

from oglio.toXmlV10.BaseOglToDom import BaseOglToDom
from oglio.toXmlV10.OglClassesToDom import OglClassesToDom
from oglio.toXmlV10.OglLinksToDom import OglLinksToDom
from oglio.toXmlV10.OglNotesToDom import OglNotesToDom
from oglio.toXmlV10.OglSequenceToDom import OglSequenceToDom
from oglio.toXmlV10.OglTextsToDom import OglTextsToDom
from oglio.toXmlV10.OglUseCasesToDom import OglUseCasesToDom
from oglio.toXmlV10.XmlConstants import XmlConstants


class OglToDom(BaseOglToDom):
    """
    The refactored version of the original methods that were part of the monolithic
     PyutXml`xxx` classes.

     This version is
        * renamed for clarity
        * uses typing for developer clarity
        * removes 'magic' strings shared between it and the ToOgl/ToPyutXml classes
        * Updated using google docstrings

    """
    VERSION:             str = '10'
    ORIGINAL_XML_PROLOG: str = '<?xml version="1.0" ?>'
    FIXED_XML_PROLOG:    str = '<?xml version="1.0" encoding="iso-8859-1"?>'

    def __init__(self, projectVersion: str, projectCodePath: str):
        """

        Args:
            projectVersion:
            projectCodePath:
        """

        self.logger:     Logger    = getLogger(__name__)

        xmlDocument, topElement = self._createStarterXmlDocument(projectVersion=projectVersion, projectCodePath=projectCodePath)

        super().__init__(xmlDocument=xmlDocument)

        self._topElement:  Element  = topElement

        self._oglClassesToMiniDom:  OglClassesToDom  = OglClassesToDom(xmlDocument=self._xmlDocument)
        self._oglLinksToMiniDom:    OglLinksToDom    = OglLinksToDom(xmlDocument=self._xmlDocument)
        self._oglUseCasesToMiniDom: OglUseCasesToDom = OglUseCasesToDom(xmlDocument=self._xmlDocument)
        self._oglSequenceToDom:   OglSequenceToDom   = OglSequenceToDom(xmlDocument=self._xmlDocument)
        self._oglNotesToDom:      OglNotesToDom      = OglNotesToDom(xmlDocument=self._xmlDocument)
        self._oglTextsToDom:      OglTextsToDom      = OglTextsToDom(xmlDocument=self._xmlDocument)

    @property
    def xmlDocument(self) -> Document:
        """
        Presumably used to persist the document

        Returns:  The serialized Document
        """
        return self._xmlDocument

    @property
    def xml(self) -> str:
        """

        Returns:  The serialized XML in pretty print format
        """
        updatedXml: str = ''
        try:
            xmlText:    str = self._xmlDocument.toxml()
            updatedXml = OglToDom.setAsISOLatin(xmlText)
        except (ValueError, Exception) as e:
            self.logger.error(f'{e=}')

        return updatedXml

    @property
    def prettyXml(self) -> str:
        updatedXml: str = ''
        try:
            xmlText:    str = self._xmlDocument.toprettyxml()
            updatedXml = OglToDom.setAsISOLatin(xmlText)
        except (ValueError, Exception) as e:
            self.logger.error(f'{e=}')

        return updatedXml

    def serialize(self, oglDocument: OglDocument):

        documentNode: Element = self._oglDocumentToXml(oglDocument=oglDocument)

        self._topElement.appendChild(documentNode)

        oglClasses:     OglClasses     = cast(OglClasses, oglDocument.oglClasses)
        oglLinks:       OglLinks       = cast(OglLinks, oglDocument.oglLinks)
        oglTexts:       OglTexts       = cast(OglTexts, oglDocument.oglTexts)
        oglUseCases:    OglUseCases    = cast(OglUseCases, oglDocument.oglUseCases)
        oglActors:      OglActors      = cast(OglActors, oglDocument.oglActors)
        oglNotes:       OglNotes       = cast(OglNotes,  oglDocument.oglNotes)
        oglSDInstances: OglSDInstances = cast(OglSDInstances, oglDocument.oglSDInstances)
        oglSDMessages:  OglSDMessages  = cast(OglSDMessages,  oglDocument.oglSDMessages)

        documentNode = self._oglClassesToMiniDom.serialize(documentNode=documentNode, oglClasses=oglClasses)
        documentNode = self._oglNotesToDom.serialize(documentNode=documentNode, oglNotes=oglNotes)
        documentNode = self._oglTextsToDom.serialize(documentNode=documentNode, oglTexts=oglTexts)
        documentNode = self._oglUseCasesToMiniDom.serialize(documentNode=documentNode, oglUseCases=oglUseCases, oglActors=oglActors)
        documentNode = self._oglLinksToMiniDom.serialize(documentNode=documentNode, oglLinks=oglLinks)

        # noinspection PyUnusedLocal
        documentNode = self._oglSequenceToDom.serialize(documentNode=documentNode, oglSDMessages=oglSDMessages, oglSDInstances=oglSDInstances)

    def writeXml(self, fqFileName: str, prettyXml: bool = True):
        """
        Persist the XML

        Args:
            fqFileName:  The fully qualified file name
            prettyXml:   Do you Barbie XML?
        """
        if prettyXml is True:
            updatedXml: str = self.prettyXml
        else:
            updatedXml = self.xml

        with open(fqFileName, 'w') as fd:
            fd.write(updatedXml)

    @classmethod
    def setAsISOLatin(cls, xmlTextToUpdate: str) -> str:
        """
        Add attribute encoding = "iso-8859-1" this is not possible with minidom, so we use pattern matching

        Args:
            xmlTextToUpdate:  Old XML

        Returns:  Updated XML
        """
        retText: str = xmlTextToUpdate.replace(OglToDom.ORIGINAL_XML_PROLOG, OglToDom.FIXED_XML_PROLOG)
        return retText

    def _createStarterXmlDocument(self, projectVersion: str, projectCodePath: str) -> Tuple[Document, Element]:

        xmlDocument: Document = Document()

        topElement: Element = xmlDocument.createElement(XmlConstants.TOP_LEVEL_ELEMENT)

        topElement.setAttribute(XmlConstants.ATTR_VERSION, projectVersion)
        topElement.setAttribute(XmlConstants.ATTR_CODE_PATH, projectCodePath)

        xmlDocument.appendChild(topElement)

        return xmlDocument, topElement

    def _oglDocumentToXml(self, oglDocument: OglDocument) -> Element:

        documentNode = self._xmlDocument.createElement(XmlConstants.ELEMENT_DOCUMENT)

        documentNode.setAttribute(XmlConstants.ATTR_TYPE, oglDocument.documentType)
        documentNode.setAttribute(XmlConstants.ATTR_TITLE, oglDocument.documentTitle)

        documentNode.setAttribute(XmlConstants.ATTR_SCROLL_POSITION_X, str(oglDocument.scrollPositionX))
        documentNode.setAttribute(XmlConstants.ATTR_SCROLL_POSITION_Y, str(oglDocument.scrollPositionY))

        documentNode.setAttribute(XmlConstants.ATTR_PIXELS_PER_UNIT_X, str(oglDocument.pixelsPerUnitX))
        documentNode.setAttribute(XmlConstants.ATTR_PIXELS_PER_UNIT_Y, str(oglDocument.pixelsPerUnitY))

        return documentNode
